@@ -1,8 +1,8 @@
 /*
  *      main.c - this file is part of Geany, a fast and lightweight IDE
  *
- *      Copyright 2005-2011 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
- *      Copyright 2006-2011 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2005-2012 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
+ *      Copyright 2006-2012 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -14,11 +14,9 @@
  *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *      GNU General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- *
- * $Id: main.c 5886 2011-08-06 18:34:14Z colombanw $
+ *      You should have received a copy of the GNU General Public License along
+ *      with this program; if not, write to the Free Software Foundation, Inc.,
+ *      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /**
@@ -45,7 +43,6 @@
 #include "main.h"
 #include "prefix.h"
 #include "prefs.h"
-#include "interface.h"
 #include "support.h"
 #include "callbacks.h"
 #include "log.h"
@@ -95,11 +92,7 @@ GeanyStatus	 main_status;
 CommandLineOptions cl_options;	/* fields initialised in parse_command_line_options */
 
 
-static const gchar geany_lib_versions[] = "GTK %u.%u.%u, GLib %u.%u.%u"
-#ifdef USE_INCLUDED_REGEX
-	", built-in regex"
-#endif
-	;
+static const gchar geany_lib_versions[] = "GTK %u.%u.%u, GLib %u.%u.%u";
 
 static gboolean want_plugins;
 
@@ -142,6 +135,7 @@ static GOptionEntry entries[] =
 	{ "no-plugins", 'p', 0, G_OPTION_ARG_NONE, &no_plugins, N_("Don't load plugins"), NULL },
 #endif
 	{ "print-prefix", 0, 0, G_OPTION_ARG_NONE, &print_prefix, N_("Print Geany's installation prefix"), NULL },
+	{ "read-only", 'r', 0, G_OPTION_ARG_NONE, &cl_options.readonly, N_("Open all FILES in read-only mode (see documention)"), NULL },
 	{ "no-session", 's', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &cl_options.load_session, N_("Don't load the previous session's files"), NULL },
 #ifdef HAVE_VTE
 	{ "no-terminal", 't', 0, G_OPTION_ARG_NONE, &no_vte, N_("Don't load terminal support"), NULL },
@@ -157,13 +151,19 @@ static GOptionEntry entries[] =
 static void setup_window_position(void)
 {
 	/* interprets the saved window geometry */
-	if (prefs.save_winpos && ui_prefs.geometry[0] != -1)
-	{
-		gtk_window_move(GTK_WINDOW(main_widgets.window), ui_prefs.geometry[0], ui_prefs.geometry[1]);
-		gtk_window_set_default_size(GTK_WINDOW(main_widgets.window), ui_prefs.geometry[2], ui_prefs.geometry[3]);
-		if (ui_prefs.geometry[4] == 1)
-			gtk_window_maximize(GTK_WINDOW(main_widgets.window));
-	}
+	if (!prefs.save_winpos)
+		return;
+
+	if (ui_prefs.geometry[0] != -1 && ui_prefs.geometry[1] != -1)
+		gtk_window_move(GTK_WINDOW(main_widgets.window),
+			ui_prefs.geometry[0], ui_prefs.geometry[1]);
+
+	if (ui_prefs.geometry[2] != -1 && ui_prefs.geometry[3] != -1)
+		gtk_window_set_default_size(GTK_WINDOW(main_widgets.window),
+			ui_prefs.geometry[2], ui_prefs.geometry[3]);
+
+	if (ui_prefs.geometry[4] == 1)
+		gtk_window_maximize(GTK_WINDOW(main_widgets.window));
 }
 
 
@@ -219,12 +219,28 @@ static void apply_settings(void)
 
 	if (interface_prefs.sidebar_pos != GTK_POS_LEFT)
 		ui_swap_sidebar_pos();
+
+	gtk_orientable_set_orientation(GTK_ORIENTABLE(ui_lookup_widget(main_widgets.window, "vpaned1")),
+		interface_prefs.msgwin_orientation);
 }
 
 
 static void main_init(void)
 {
+	/* add our icon path in case we aren't installed in the system prefix */
+#ifndef G_OS_WIN32
+	gchar *path = g_build_filename(GEANY_DATADIR, "icons", NULL);
+	gtk_icon_theme_append_search_path(gtk_icon_theme_get_default(), path);
+	g_free(path);
+#else
+	gtk_icon_theme_append_search_path(gtk_icon_theme_get_default(), "share\\icons");
+#endif
+
 	/* inits */
+	ui_init_stock_items();
+
+	ui_init_builder();
+
 	main_widgets.window				= NULL;
 	app->project			= NULL;
 	ui_widgets.open_fontsel		= NULL;
@@ -240,8 +256,6 @@ static void main_init(void)
 	ui_prefs.recent_projects_queue		= g_queue_new();
 	main_status.opening_session_files	= FALSE;
 
-	ui_init_stock_items();
-
 	main_widgets.window = create_window1();
 
 	/* add recent projects to the Project menu */
@@ -269,9 +283,6 @@ static void main_init(void)
 	gtk_widget_set_name(ui_lookup_widget(main_widgets.window, "menubar1"), "GeanyMenubar");
 	gtk_widget_set_name(main_widgets.toolbar, "GeanyToolbar");
 
-	/* for some reason we need to set the initial size request,
-	 * otherwise the main window gets crazy dimensions */
-	gtk_widget_set_size_request(main_widgets.window, 10, 10);
 	gtk_window_set_default_size(GTK_WINDOW(main_widgets.window),
 		GEANY_WINDOW_DEFAULT_WIDTH, GEANY_WINDOW_DEFAULT_HEIGHT);
 }
@@ -279,7 +290,7 @@ static void main_init(void)
 
 const gchar *main_get_version_string(void)
 {
-	static gchar full[] = VERSION " (svn >= " REVISION ")";
+	static gchar full[] = VERSION " (git >= " REVISION ")";
 
 	if (utils_str_equal(REVISION, "-1"))
 		return VERSION;
@@ -365,6 +376,18 @@ static void get_line_and_column_from_filename(gchar *filename, gint *line, gint
 }
 
 
+#ifdef G_OS_WIN32
+static void change_working_directory_on_windows(const gchar *install_dir)
+{
+	/* On Windows, change the working directory to the Geany installation path to not lock
+	 * the directory of a file passed as command line argument (see bug #2626124).
+	 * This also helps if plugins or other code uses relative paths to load
+	 * any additional resources (e.g. share/geany-plugins/...). */
+	win32_set_working_directory(install_dir);
+}
+#endif
+
+
 static void setup_paths(void)
 {
 	gchar *data_dir;
@@ -376,13 +399,15 @@ static void setup_paths(void)
 	 * documentation and data files */
 	gchar *install_dir = win32_get_installation_dir();
 
-	data_dir = g_strconcat(install_dir, "\\data", NULL); /* e.g. C:\Program Files\geany\data */
-	doc_dir = g_strconcat(install_dir, "\\doc", NULL);
+	data_dir = g_build_filename(install_dir, "data", NULL); /* e.g. C:\Program Files\geany\data */
+	doc_dir = g_build_filename(install_dir, "doc", NULL);
+
+	change_working_directory_on_windows(install_dir);
 
 	g_free(install_dir);
 #else
-	data_dir = g_strconcat(GEANY_DATADIR, "/geany", NULL); /* e.g. /usr/share/geany */
-	doc_dir = g_strconcat(GEANY_DOCDIR, "/html", NULL);
+	data_dir = g_build_filename(GEANY_DATADIR, "geany", NULL); /* e.g. /usr/share/geany */
+	doc_dir = g_build_filename(GEANY_DOCDIR, "html", NULL);
 #endif
 
 	/* convert path names to locale encoding */
@@ -450,7 +475,7 @@ void main_locale_init(const gchar *locale_dir, const gchar *package)
 	{
 		gchar *install_dir = win32_get_installation_dir();
 		/* e.g. C:\Program Files\geany\lib\locale */
-		l_locale_dir = g_strconcat(install_dir, "\\share\\locale", NULL);
+		l_locale_dir = g_build_filename(install_dir, "share", "locale", NULL);
 		g_free(install_dir);
 	}
 #else
@@ -481,12 +506,24 @@ static void print_filetypes(void)
 }
 
 
+static void wait_for_input_on_windows(void)
+{
+#ifdef G_OS_WIN32
+	if (verbose_mode)
+	{
+		geany_debug("Press any key to continue");
+		getchar();
+	}
+#endif
+}
+
+
 static void parse_command_line_options(gint *argc, gchar ***argv)
 {
 	GError *error = NULL;
 	GOptionContext *context;
 	gint i;
-	CommandLineOptions def_clo = {FALSE, NULL, TRUE, -1, -1, FALSE, FALSE};
+	CommandLineOptions def_clo = {FALSE, NULL, TRUE, -1, -1, FALSE, FALSE, FALSE};
 
 	/* first initialise cl_options fields with default values */
 	cl_options = def_clo;
@@ -516,6 +553,19 @@ static void parse_command_line_options(gint *argc, gchar ***argv)
 		exit(1);
 	}
 
+	app->debug_mode = verbose_mode;
+	if (app->debug_mode)
+	{
+		/* Since GLib 2.32 messages logged with levels INFO and DEBUG aren't output by the
+		 * default log handler unless the G_MESSAGES_DEBUG environment variable contains the
+		 * domain of the message or is set to the special value "all" */
+		g_setenv("G_MESSAGES_DEBUG", "all", FALSE);
+	}
+
+#ifdef G_OS_WIN32
+	win32_init_debug_code();
+#endif
+
 	if (show_version)
 	{
 		printf(PACKAGE " %s (", main_get_version_string());
@@ -525,6 +575,7 @@ static void parse_command_line_options(gint *argc, gchar ***argv)
 			GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION,
 			GLIB_MAJOR_VERSION, GLIB_MINOR_VERSION, GLIB_MICRO_VERSION);
 		printf(")\n");
+		wait_for_input_on_windows();
 		exit(0);
 	}
 
@@ -534,15 +585,10 @@ static void parse_command_line_options(gint *argc, gchar ***argv)
 		printf("%s\n", GEANY_DATADIR);
 		printf("%s\n", GEANY_LIBDIR);
 		printf("%s\n", GEANY_LOCALEDIR);
+		wait_for_input_on_windows();
 		exit(0);
 	}
 
-	app->debug_mode = verbose_mode;
-
-#ifdef G_OS_WIN32
-	win32_init_debug_code();
-#endif
-
 	if (alternate_config)
 	{
 		geany_debug("alternate config: %s", alternate_config);
@@ -560,12 +606,14 @@ static void parse_command_line_options(gint *argc, gchar ***argv)
 		filetypes_init_types();
 		ret = symbols_generate_global_tags(*argc, *argv, ! no_preprocessing);
 		filetypes_free_types();
+		wait_for_input_on_windows();
 		exit(ret);
 	}
 
 	if (ft_names)
 	{
 		print_filetypes();
+		wait_for_input_on_windows();
 		exit(0);
 	}
 
@@ -706,7 +754,7 @@ static gint setup_config_dir(void)
 	gint mkdir_result = 0;
 
 	/* convert configdir to locale encoding to avoid troubles */
-	setptr(app->configdir, utils_get_locale_from_utf8(app->configdir));
+	SETPTR(app->configdir, utils_get_locale_from_utf8(app->configdir));
 
 	mkdir_result = create_config_dir();
 	if (mkdir_result != 0)
@@ -721,12 +769,13 @@ static gint setup_config_dir(void)
 	}
 	/* make configdir a real path */
 	if (g_file_test(app->configdir, G_FILE_TEST_EXISTS))
-		setptr(app->configdir, tm_get_real_path(app->configdir));
+		SETPTR(app->configdir, tm_get_real_path(app->configdir));
 
 	return mkdir_result;
 }
 
-
+/* Signal handling removed since on_exit_clicked() uses functions that are
+ * illegal in signal handlers
 static void signal_cb(gint sig)
 {
 	if (sig == SIGTERM)
@@ -734,7 +783,7 @@ static void signal_cb(gint sig)
 		on_exit_clicked(NULL, NULL);
 	}
 }
-
+ */
 
 /* Used for command-line arguments at startup or from socket.
  * this will strip any :line:col filename suffix from locale_filename */
@@ -759,10 +808,10 @@ gboolean main_handle_filename(const gchar *locale_filename)
 
 	if (g_file_test(filename, G_FILE_TEST_IS_REGULAR))
 	{
-		doc = document_open_file(filename, FALSE, NULL, NULL);
+		doc = document_open_file(filename, cl_options.readonly, NULL, NULL);
 		/* add recent file manually if opening_session_files is set */
 		if (doc != NULL && main_status.opening_session_files)
-			ui_add_recent_file(doc->file_name);
+			ui_add_recent_document(doc);
 		g_free(filename);
 		return TRUE;
 	}
@@ -772,7 +821,7 @@ gboolean main_handle_filename(const gchar *locale_filename)
 
 		doc = document_new_file(utf8_filename, NULL, NULL);
 		if (doc != NULL)
-			ui_add_recent_file(doc->file_name);
+			ui_add_recent_document(doc);
 		g_free(utf8_filename);
 		g_free(filename);
 		return TRUE;
@@ -783,12 +832,10 @@ gboolean main_handle_filename(const gchar *locale_filename)
 
 
 /* open files from command line */
-static gboolean open_cl_files(gint argc, gchar **argv)
+static void open_cl_files(gint argc, gchar **argv)
 {
 	gint i;
 
-	if (argc <= 1) return FALSE;
-
 	for (i = 1; i < argc; i++)
 	{
 		gchar *filename = main_get_argv_filename(argv[i]);
@@ -801,7 +848,7 @@ static gboolean open_cl_files(gint argc, gchar **argv)
 
 #ifdef G_OS_WIN32
 		/* It seems argv elements are encoded in CP1252 on a German Windows */
-		setptr(filename, g_locale_to_utf8(filename, -1, NULL, NULL, NULL));
+		SETPTR(filename, g_locale_to_utf8(filename, -1, NULL, NULL, NULL));
 #endif
 		if (filename && ! main_handle_filename(filename))
 		{
@@ -813,7 +860,6 @@ static gboolean open_cl_files(gint argc, gchar **argv)
 		}
 		g_free(filename);
 	}
-	return TRUE;
 }
 
 
@@ -851,13 +897,9 @@ static void load_settings(void)
 
 void main_load_project_from_command_line(const gchar *locale_filename, gboolean use_session)
 {
-	gchar *pfile = NULL;
-
-	if (utils_is_uri(locale_filename))
-		pfile = utils_get_path_from_uri(locale_filename);
-	else
-		pfile = g_strdup(locale_filename);
+	gchar *pfile;
 
+	pfile = utils_get_path_from_uri(locale_filename);
 	if (pfile != NULL)
 	{
 		if (use_session)
@@ -871,39 +913,42 @@ void main_load_project_from_command_line(const gchar *locale_filename, gboolean
 
 static void load_startup_files(gint argc, gchar **argv)
 {
-	gboolean load_project_from_cl = FALSE;
+	gboolean load_session = FALSE;
 
-	/* ATM when opening a project file any other filenames are ignored */
-	load_project_from_cl = (argc > 1) && g_str_has_suffix(argv[1], ".geany");
-	if (load_project_from_cl && argc > 2)
-		g_print("Ignoring extra filenames after %s", argv[1]);
-
-	if (load_project_from_cl || ! open_cl_files(argc, argv))
+	if (argc > 1 && g_str_has_suffix(argv[1], ".geany"))
 	{
-		if (prefs.load_session)
+		/* project file specified: load it, but decide the session later */
+		main_load_project_from_command_line(argv[1], FALSE);
+		argc--, argv++;
+		/* force session load if using project-based session files */
+		load_session = project_prefs.project_session;
+	}
+
+	/* Load the default session if:
+	 * 1. "Load files from the last session" is active.
+	 * 2. --no-session is not specified.
+	 * 3. We are a primary instance.
+	 * Has no effect if a CL project is loaded and using project-based session files. */
+	if (prefs.load_session && cl_options.load_session && !cl_options.new_instance)
+	{
+		if (app->project == NULL)
+			load_session_project_file();
+		load_session = TRUE;
+	}
+
+	if (load_session)
+	{
+		/* load session files into tabs, as they are found in the session_files variable */
+		configuration_open_files();
+
+		if (gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) == 0)
 		{
-			if (load_project_from_cl)
-			{
-				main_load_project_from_command_line(argv[1], FALSE);
-			}
-			else if (cl_options.load_session && !cl_options.new_instance)
-				load_session_project_file();
-
-			/* when we want a new instance, we still load project session files unless -s
-			 * was passed */
-			if (!cl_options.load_session || (!load_project_from_cl && cl_options.new_instance))
-				return;
-
-			/* load session files into tabs, as they are found in the session_files variable */
-			configuration_open_files();
-
-			if (gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) == 0)
-			{
-				ui_update_popup_copy_items(NULL);
-				ui_update_popup_reundo_items(NULL);
-			}
+			ui_update_popup_copy_items(NULL);
+			ui_update_popup_reundo_items(NULL);
 		}
 	}
+
+	open_cl_files(argc, argv);
 }
 
 
@@ -954,8 +999,8 @@ gint main(gint argc, gchar **argv)
 	 * dependencies (e.g. WebKit, Soup, ...) */
 	if (!g_thread_supported())
 		g_thread_init(NULL);
-
-	signal(SIGTERM, signal_cb);
+    /* removed as signal handling was wrong, see signal_cb()
+	signal(SIGTERM, signal_cb); */
 #ifdef G_OS_UNIX
 	/* SIGQUIT is used to kill spawned children and we get also this signal, so ignore */
 	signal(SIGQUIT, SIG_IGN);
@@ -1022,7 +1067,6 @@ gint main(gint argc, gchar **argv)
 	sidebar_init();
 	load_settings();	/* load keyfile */
 
-	highlighting_init();
 	msgwin_init();
 	build_init();
 	ui_create_insert_menu_items();
@@ -1036,13 +1080,6 @@ gint main(gint argc, gchar **argv)
 	symbols_init();
 	editor_snippets_init();
 
-	/* set window icon */
-	{
-		GdkPixbuf *pb = ui_new_pixbuf_from_inline(GEANY_IMAGE_LOGO);
-		gtk_window_set_icon(GTK_WINDOW(main_widgets.window), pb);
-		g_object_unref(pb);	/* free our reference */
-	}
-
 	/* registering some basic events */
 	g_signal_connect(main_widgets.window, "delete-event", G_CALLBACK(on_exit_clicked), NULL);
 	g_signal_connect(main_widgets.window, "window-state-event", G_CALLBACK(on_window_state_event), NULL);
@@ -1092,7 +1129,6 @@ gint main(gint argc, gchar **argv)
 	ui_save_buttons_toggle(FALSE);
 
 	doc = document_get_current();
-	gtk_widget_grab_focus(GTK_WIDGET(doc->editor->sci));
 	sidebar_select_openfiles_item(doc);
 	build_menu_update(doc);
 	sidebar_update_tag_list(doc, FALSE);
@@ -1105,6 +1141,7 @@ gint main(gint argc, gchar **argv)
 	setup_window_position();
 
 	/* finally show the window */
+	document_grab_focus(doc);
 	gtk_widget_show(main_widgets.window);
 	main_status.main_window_realized = TRUE;
 
@@ -1120,17 +1157,6 @@ gint main(gint argc, gchar **argv)
 	}
 #endif
 
-#ifdef G_OS_WIN32
-	{
-		gchar *dir;
-		/* On Windows, change the working directory to the Geany installation path to not lock
-		 * the directory of a file passed as command line argument (see bug #2626124). */
-		dir = win32_get_installation_dir();
-		win32_set_working_directory(dir);
-		g_free(dir);
-	}
-#endif
-
 	/* when we are really done with setting everything up and the main event loop is running,
 	 * tell other components, mainly plugins, that startup is complete */
 	g_idle_add_full(G_PRIORITY_LOW, send_startup_complete, NULL, NULL);
@@ -1164,6 +1190,7 @@ void main_quit()
 
 	navqueue_free();
 	keybindings_free();
+	notebook_free();
 	highlighting_free_styles();
 	templates_free_templates();
 	msgwin_finalize();
@@ -1207,6 +1234,7 @@ void main_quit()
 	g_free(printing_prefs.external_print_cmd);
 	g_free(printing_prefs.page_header_datefmt);
 	g_strfreev(ui_prefs.custom_commands);
+	g_strfreev(ui_prefs.custom_commands_labels);
 
 	queue_free(ui_prefs.recent_queue);
 	queue_free(ui_prefs.recent_projects_queue);
@@ -1238,6 +1266,8 @@ void main_quit()
 
 	g_free(app);
 
+	ui_finalize_builder();
+
 	gtk_main_quit();
 }
 
