@@ -1,8 +1,9 @@
 /*
  *      ui_utils.c - this file is part of Geany, a fast and lightweight IDE
  *
- *      Copyright 2006-2011 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
- *      Copyright 2006-2011 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2006-2012 Enrico Tröger <enrico(dot)troeger(at)uvena(dot)de>
+ *      Copyright 2006-2012 Nick Treleaven <nick(dot)treleaven(at)btinternet(dot)com>
+ *      Copyright 2011-2012 Matthew Brush <mbrush(at)codebrainz(dot)ca>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -14,11 +15,9 @@
  *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *      GNU General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- *
- * $Id: ui_utils.c 5944 2011-09-22 11:52:58Z ntrel $
+ *      You should have received a copy of the GNU General Public License along
+ *      with this program; if not, write to the Free Software Foundation, Inc.,
+ *      51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /** @file ui_utils.h
@@ -27,11 +26,14 @@
 
 #include "geany.h"
 
+#include "support.h"
+
 #include <string.h>
 #include <ctype.h>
 #include <gdk/gdkkeysyms.h>
 
 #include "ui_utils.h"
+#include "dialogs.h"
 #include "prefs.h"
 #include "sciwrappers.h"
 #include "document.h"
@@ -42,7 +44,6 @@
 #include "utils.h"
 #include "callbacks.h"
 #include "encodings.h"
-#include "images.c"
 #include "sidebar.h"
 #include "win32.h"
 #include "project.h"
@@ -62,6 +63,13 @@ GeanyMainWidgets	main_widgets;
 UIPrefs			ui_prefs;
 UIWidgets		ui_widgets;
 
+static GtkBuilder *builder = NULL;
+static GtkWidget* window1 = NULL;
+static GtkWidget* toolbar_popup_menu1 = NULL;
+static GtkWidget* edit_menu1 = NULL;
+static GtkWidget* prefs_dialog = NULL;
+static GtkWidget* project_dialog = NULL;
+
 static struct
 {
 	/* pointers to widgets only sensitive when there is at least one document, the pointers can
@@ -113,6 +121,7 @@ void ui_widget_set_sensitive(GtkWidget *widget, gboolean set)
  * that didn't use allow_override and has not timed out. */
 static void set_statusbar(const gchar *text, gboolean allow_override)
 {
+	static guint id = 0;
 	static glong last_time = 0;
 	GTimeVal timeval;
 	const gint GEANY_STATUS_TIMEOUT = 1;
@@ -120,19 +129,22 @@ static void set_statusbar(const gchar *text, gboolean allow_override)
 	if (! interface_prefs.statusbar_visible)
 		return; /* just do nothing if statusbar is not visible */
 
+	if (id == 0)
+		id = gtk_statusbar_get_context_id(GTK_STATUSBAR(ui_widgets.statusbar), "geany-main");
+
 	g_get_current_time(&timeval);
 
 	if (! allow_override)
 	{
-		gtk_statusbar_pop(GTK_STATUSBAR(ui_widgets.statusbar), 1);
-		gtk_statusbar_push(GTK_STATUSBAR(ui_widgets.statusbar), 1, text);
+		gtk_statusbar_pop(GTK_STATUSBAR(ui_widgets.statusbar), id);
+		gtk_statusbar_push(GTK_STATUSBAR(ui_widgets.statusbar), id, text);
 		last_time = timeval.tv_sec;
 	}
 	else
 	if (timeval.tv_sec > last_time + GEANY_STATUS_TIMEOUT)
 	{
-		gtk_statusbar_pop(GTK_STATUSBAR(ui_widgets.statusbar), 1);
-		gtk_statusbar_push(GTK_STATUSBAR(ui_widgets.statusbar), 1, text);
+		gtk_statusbar_pop(GTK_STATUSBAR(ui_widgets.statusbar), id);
+		gtk_statusbar_push(GTK_STATUSBAR(ui_widgets.statusbar), id, text);
 	}
 }
 
@@ -169,6 +181,7 @@ static void add_statusbar_statistics(GString *stats_str,
 	const gchar *fmt;
 	const gchar *expos;	/* % expansion position */
 	const gchar sp[] = "      ";
+	ScintillaObject *sci = doc->editor->sci;
 
 	fmt = NZV(statusbar_template) ? statusbar_template :
 		/* Status bar statistics: col = column, sel = selection. */
@@ -197,9 +210,19 @@ static void add_statusbar_statistics(GString *stats_str,
 				g_string_append_printf(stats_str, "%d", col + 1);
 				break;
 			case 's':
-				g_string_append_printf(stats_str, "%d",
-					sci_get_selected_text_length(doc->editor->sci) - 1);
+			{
+				gint len = sci_get_selected_text_length(sci) - 1;
+				/* check if whole lines are selected */
+				if (!len || sci_get_col_from_position(sci,
+						sci_get_selection_start(sci)) != 0 ||
+					sci_get_col_from_position(sci,
+						sci_get_selection_end(sci)) != 0)
+					g_string_append_printf(stats_str, "%d", len);
+				else /* L = lines */
+					g_string_append_printf(stats_str, _("%dL"),
+						sci_get_lines_selected(doc->editor->sci) - 1);
 				break;
+			}
 			case 'w':
 				/* RO = read-only */
 				g_string_append(stats_str, (doc->readonly) ? _("RO ") :
@@ -252,7 +275,7 @@ static void add_statusbar_statistics(GString *stats_str,
 				g_string_append(stats_str, filetypes_get_display_name(doc->file_type));
 				break;
 			case 'S':
-				symbols_get_current_function(doc, &cur_tag);
+				symbols_get_current_scope(doc, &cur_tag);
 				g_string_append(stats_str, cur_tag);
 				break;
 			default:
@@ -305,9 +328,9 @@ void ui_update_statusbar(GeanyDocument *doc, gint pos)
 	{
 		const gchar sp[] = "      ";
 		g_string_append(stats_str, sp);
-		g_string_append_printf(stats_str, "pos: %d", pos);
+		g_string_append_printf(stats_str, _("pos: %d"), pos);
 		g_string_append(stats_str, sp);
-		g_string_append_printf(stats_str, "style: %d", sci_get_style_at(doc->editor->sci, pos));
+		g_string_append_printf(stats_str, _("style: %d"), sci_get_style_at(doc->editor->sci, pos));
 	}
 #endif
 		/* can be overridden by status messages */
@@ -661,7 +684,7 @@ void ui_save_buttons_toggle(gboolean enable)
 	gboolean dirty_tabs = FALSE;
 
 	if (ui_prefs.allow_always_save)
-		enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) > 0 ? TRUE : FALSE;
+		enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) > 0;
 
 	ui_widget_set_sensitive(widgets.save_buttons[0], enable);
 	ui_widget_set_sensitive(widgets.save_buttons[1], enable);
@@ -744,8 +767,6 @@ static void init_document_widgets(void)
 	add_doc_toolitem("CloseAll");
 	add_doc_toolitem("Search");
 	add_doc_toolitem("SearchEntry");
-	add_doc_toolitem("NavBack");
-	add_doc_toolitem("NavFor");
 	add_doc_toolitem("ZoomIn");
 	add_doc_toolitem("ZoomOut");
 	add_doc_toolitem("Indent");
@@ -771,7 +792,7 @@ static void init_document_widgets(void)
 void ui_document_buttons_update(void)
 {
 	guint i;
-	gboolean enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) ? TRUE : FALSE;
+	gboolean enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) > 0;
 
 	for (i = 0; i < widgets.document_buttons->len; i++)
 	{
@@ -799,7 +820,7 @@ static void on_doc_sensitive_widget_destroy(GtkWidget *widget, G_GNUC_UNUSED gpo
  **/
 void ui_add_document_sensitive(GtkWidget *widget)
 {
-	gboolean enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) ? TRUE : FALSE;
+	gboolean enable = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook)) > 0;
 
 	ui_widget_set_sensitive(widget, enable);
 
@@ -926,101 +947,7 @@ void ui_document_show_hide(GeanyDocument *doc)
 
 void ui_set_search_entry_background(GtkWidget *widget, gboolean success)
 {
-	static const GdkColor red   = {0, 0xffff, 0x6666, 0x6666};
-	static const GdkColor white = {0, 0xffff, 0xffff, 0xffff};
-	static gboolean old_value = TRUE;
-
-	g_return_if_fail(widget != NULL);
-
-	/* update only if really needed */
-	if (old_value != success)
-	{
-		gtk_widget_modify_base(widget, GTK_STATE_NORMAL, success ? NULL : &red);
-		gtk_widget_modify_text(widget, GTK_STATE_NORMAL, success ? NULL : &white);
-
-		old_value = success;
-	}
-}
-
-
-static gboolean have_tango_icon_theme(void)
-{
-	static gboolean result = FALSE;
-	static gboolean checked = FALSE;
-
-	if (! checked)
-	{
-		gchar *theme_name;
-
-		g_object_get(G_OBJECT(gtk_settings_get_default()), "gtk-icon-theme-name", &theme_name, NULL);
-		setptr(theme_name, g_utf8_strdown(theme_name, -1));
-
-		result = (strstr(theme_name, "tango") != NULL);
-		checked = TRUE;
-
-		g_free(theme_name);
-	}
-
-	return result;
-}
-
-
-/* Note: remember to unref the pixbuf once an image or window has added a reference. */
-GdkPixbuf *ui_new_pixbuf_from_inline(gint img)
-{
-	switch (img)
-	{
-		case GEANY_IMAGE_LOGO:
-			return gdk_pixbuf_new_from_inline(-1, aladin_inline, FALSE, NULL);
-			break;
-		case GEANY_IMAGE_SAVE_ALL:
-		{
-			/* check whether the icon theme looks like a Gnome icon theme, if so use the
-			 * old Gnome based Save All icon, otherwise assume a Tango-like icon theme */
-			if (have_tango_icon_theme())
-				return gdk_pixbuf_new_from_inline(-1, save_all_tango_inline, FALSE, NULL);
-			else
-				return gdk_pixbuf_new_from_inline(-1, save_all_gnome_inline, FALSE, NULL);
-			break;
-		}
-		case GEANY_IMAGE_CLOSE_ALL:
-		{
-			return gdk_pixbuf_new_from_inline(-1, close_all_inline, FALSE, NULL);
-			break;
-		}
-		case GEANY_IMAGE_BUILD:
-		{
-			return gdk_pixbuf_new_from_inline(-1, build_inline, FALSE, NULL);
-			break;
-		}
-		default:
-			return NULL;
-	}
-}
-
-
-static GdkPixbuf *ui_new_pixbuf_from_stock(const gchar *stock_id)
-{
-	if (utils_str_equal(stock_id, GEANY_STOCK_CLOSE_ALL))
-		return ui_new_pixbuf_from_inline(GEANY_IMAGE_CLOSE_ALL);
-	else if (utils_str_equal(stock_id, GEANY_STOCK_BUILD))
-		return ui_new_pixbuf_from_inline(GEANY_IMAGE_BUILD);
-	else if (utils_str_equal(stock_id, GEANY_STOCK_SAVE_ALL))
-		return ui_new_pixbuf_from_inline(GEANY_IMAGE_SAVE_ALL);
-
-	return NULL;
-}
-
-
-GtkWidget *ui_new_image_from_inline(gint img)
-{
-	GtkWidget *wid;
-	GdkPixbuf *pb;
-
-	pb = ui_new_pixbuf_from_inline(img);
-	wid = gtk_image_new_from_pixbuf(pb);
-	g_object_unref(pb);	/* the image doesn't adopt our reference, so remove our ref. */
-	return wid;
+	gtk_widget_set_name(widget, success ? NULL : "geany-search-entry-no-match");
 }
 
 
@@ -1115,22 +1042,23 @@ static void recent_project_activate_cb(GtkMenuItem *menuitem, G_GNUC_UNUSED gpoi
 }
 
 
-static void add_recent_file(const gchar *utf8_filename, GeanyRecentFiles *grf)
+static void add_recent_file(const gchar *utf8_filename, GeanyRecentFiles *grf,
+		const GtkRecentData *rdata)
 {
 	if (g_queue_find_custom(grf->recent_queue, utf8_filename, (GCompareFunc) strcmp) == NULL)
 	{
-#if GTK_CHECK_VERSION(2, 10, 0)
-		if (grf->type == RECENT_FILE_FILE)
+
+		if (grf->type == RECENT_FILE_FILE && rdata)
 		{
 			GtkRecentManager *manager = gtk_recent_manager_get_default();
 			gchar *uri = g_filename_to_uri(utf8_filename, NULL, NULL);
 			if (uri != NULL)
 			{
-				gtk_recent_manager_add_item(manager, uri);
+				gtk_recent_manager_add_full(manager, uri, rdata);
 				g_free(uri);
 			}
 		}
-#endif
+
 		g_queue_push_head(grf->recent_queue, g_strdup(utf8_filename));
 		if (g_queue_get_length(grf->recent_queue) > file_prefs.mru_length)
 		{
@@ -1144,15 +1072,34 @@ static void add_recent_file(const gchar *utf8_filename, GeanyRecentFiles *grf)
 }
 
 
-void ui_add_recent_file(const gchar *utf8_filename)
+void ui_add_recent_document(GeanyDocument *doc)
 {
-	add_recent_file(utf8_filename, recent_get_recent_files());
+	/* what are the groups for actually? */
+	static const gchar *groups[2] = {
+		"geany",
+		NULL
+	};
+	GtkRecentData rdata;
+
+	/* Prepare the data for gtk_recent_manager_add_full() */
+	rdata.display_name = NULL;
+	rdata.description = NULL;
+	rdata.mime_type = doc->file_type->mime_type;
+	/* if we ain't got no mime-type, fallback to plain text */
+	if (! rdata.mime_type)
+		rdata.mime_type = (gchar *) "text/plain";
+	rdata.app_name = (gchar *) "geany";
+	rdata.app_exec = (gchar *) "geany %u";
+	rdata.groups = (gchar **) groups;
+	rdata.is_private = FALSE;
+
+	add_recent_file(doc->file_name, recent_get_recent_files(), &rdata);
 }
 
 
 void ui_add_recent_project_file(const gchar *utf8_filename)
 {
-	add_recent_file(utf8_filename, recent_get_recent_projects());
+	add_recent_file(utf8_filename, recent_get_recent_projects(), NULL);
 }
 
 
@@ -1161,9 +1108,9 @@ gchar *ui_menu_item_get_text(GtkMenuItem *menu_item)
 {
 	const gchar *text = NULL;
 
-	if (GTK_BIN(menu_item)->child)
+	if (gtk_bin_get_child(GTK_BIN(menu_item)))
 	{
-		GtkWidget *child = GTK_BIN(menu_item)->child;
+		GtkWidget *child = gtk_bin_get_child(GTK_BIN(menu_item));
 
 		if (GTK_IS_LABEL(child))
 			text = gtk_label_get_text(GTK_LABEL(child));
@@ -1294,12 +1241,11 @@ static void update_recent_menu(GeanyRecentFiles *grf)
 
 void ui_toggle_editor_features(GeanyUIEditorFeatures feature)
 {
-	gint i, max = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
-	GeanyDocument *doc;
+	guint i;
 
-	for (i = 0; i < max; i++)
+	foreach_document (i)
 	{
-		doc = document_get_from_page(i);
+		GeanyDocument *doc = documents[i];
 
 		switch (feature)
 		{
@@ -1366,7 +1312,7 @@ GtkWidget *ui_dialog_vbox_new(GtkDialog *dialog)
 	GtkWidget *vbox = gtk_vbox_new(FALSE, 12);	/* need child vbox to set a separate border. */
 
 	gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);
-	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);
+	gtk_container_add(GTK_CONTAINER(gtk_dialog_get_content_area(GTK_DIALOG(dialog))), vbox);
 	return vbox;
 }
 
@@ -1423,21 +1369,15 @@ static void entry_clear_icon_release_cb(GtkEntry *entry, gint icon_pos,
 /** Adds a small clear icon to the right end of the passed @a entry.
  *  A callback to clear the contents of the GtkEntry is automatically added.
  *
- *  This feature is only available with GTK 2.16 but implemented as a runtime check,
- *  so it is safe to just use this function, if the code is ran with older versions,
- *  nothing happens. If ran with GTK 2.16 or newer, the icon is displayed.
- *
  * @param entry The GtkEntry object to which the icon should be attached.
  *
  *  @since 0.16
  */
 void ui_entry_add_clear_icon(GtkEntry *entry)
 {
-	if (gtk_check_version(2, 15, 2) == NULL)
-	{
-		g_object_set(entry, "secondary-icon-stock", "gtk-clear", NULL);
-		g_signal_connect(entry, "icon-release", G_CALLBACK(entry_clear_icon_release_cb), NULL);
-	}
+	g_object_set(entry, "secondary-icon-stock", GTK_STOCK_CLEAR,
+		"secondary-icon-activatable", TRUE, NULL);
+	g_signal_connect(entry, "icon-release", G_CALLBACK(entry_clear_icon_release_cb), NULL);
 }
 
 
@@ -1536,7 +1476,7 @@ void ui_combo_box_add_to_history(GtkComboBoxEntry *combo_entry,
 	if (history_len <= 0)
 		history_len = 10;
 	if (!text)
-		text = gtk_entry_get_text(GTK_ENTRY(GTK_BIN(combo)->child));
+		text = gtk_entry_get_text(GTK_ENTRY(gtk_bin_get_child(GTK_BIN(combo))));
 
 	model = gtk_combo_box_get_model(combo);
 
@@ -1575,11 +1515,7 @@ void ui_combo_box_prepend_text_once(GtkComboBox *combo, const gchar *text)
  * document status. */
 void ui_update_tab_status(GeanyDocument *doc)
 {
-	const GdkColor *color = document_get_status_color(doc);
-
-	/* NULL color will reset to default */
-	gtk_widget_modify_fg(doc->priv->tab_label, GTK_STATE_NORMAL, color);
-	gtk_widget_modify_fg(doc->priv->tab_label, GTK_STATE_ACTIVE, color);
+	gtk_widget_set_name(doc->priv->tab_label, document_get_status_widget_class(doc));
 
 	sidebar_openfiles_update(doc);
 }
@@ -1727,7 +1663,7 @@ void ui_setup_open_button_callback(GtkWidget *open_btn, const gchar *title,
 	if (title)
 		g_object_set_data_full(G_OBJECT(open_btn), "title", g_strdup(title),
 				(GDestroyNotify) g_free);
-	g_object_set_data(G_OBJECT(open_btn), "action", (gpointer) action);
+	g_object_set_data(G_OBJECT(open_btn), "action", GINT_TO_POINTER(action));
 	ui_hookup_widget(open_btn, path_entry, "entry");
 	g_signal_connect(open_btn, "clicked", G_CALLBACK(ui_path_box_open_clicked), open_btn);
 }
@@ -1775,10 +1711,8 @@ static gchar *run_file_chooser(const gchar *title, GtkFileChooserAction action,
 static void ui_path_box_open_clicked(GtkButton *button, gpointer user_data)
 {
 	GtkWidget *path_box = GTK_WIDGET(user_data);
-	GtkFileChooserAction action =
-		(GtkFileChooserAction) g_object_get_data(G_OBJECT(path_box), "action");
-	GtkEntry *entry =
-		(GtkEntry *) g_object_get_data(G_OBJECT(path_box), "entry");
+	GtkFileChooserAction action = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(path_box), "action"));
+	GtkEntry *entry = g_object_get_data(G_OBJECT(path_box), "entry");
 	const gchar *title = g_object_get_data(G_OBJECT(path_box), "title");
 	gchar *utf8_path = NULL;
 
@@ -1841,7 +1775,7 @@ void ui_statusbar_showhide(gboolean state)
 void ui_table_add_row(GtkTable *table, gint row, ...)
 {
 	va_list args;
-	gint i;
+	guint i;
 	GtkWidget *widget;
 
 	va_start(args, row);
@@ -1885,6 +1819,12 @@ static void on_config_file_clicked(GtkWidget *widget, gpointer user_data)
 }
 
 
+static void free_on_closure_notify(gpointer data, GClosure *closure)
+{
+	g_free(data);
+}
+
+
 /* @note You should connect to the "document-save" signal yourself to detect
  * if the user has just saved the config file, reloading it. */
 void ui_add_config_file_menu_item(const gchar *real_path, const gchar *label, GtkContainer *parent)
@@ -1907,9 +1847,8 @@ void ui_add_config_file_menu_item(const gchar *real_path, const gchar *label, Gt
 
 	gtk_widget_show(item);
 	gtk_container_add(parent, item);
-	g_signal_connect(item, "activate", G_CALLBACK(on_config_file_clicked),
-		/* this memory is kept */
-		g_strdup(real_path));
+	g_signal_connect_data(item, "activate", G_CALLBACK(on_config_file_clicked),
+			g_strdup(real_path), free_on_closure_notify, 0);
 }
 
 
@@ -1936,10 +1875,6 @@ static void create_config_files_menu(void)
 
 void ui_init_stock_items(void)
 {
-	GtkIconSet *icon_set;
-	GtkIconFactory *factory = gtk_icon_factory_new();
-	GdkPixbuf *pb;
-	gsize i, len;
 	GtkStockItem items[] =
 	{
 		{ GEANY_STOCK_SAVE_ALL, N_("Save All"), 0, 0, GETTEXT_PACKAGE },
@@ -1947,20 +1882,7 @@ void ui_init_stock_items(void)
 		{ GEANY_STOCK_BUILD, N_("Build"), 0, 0, GETTEXT_PACKAGE }
 	};
 
-	len = G_N_ELEMENTS(items);
-	for (i = 0; i < len; i++)
-	{
-		pb = ui_new_pixbuf_from_stock(items[i].stock_id);
-		icon_set = gtk_icon_set_new_from_pixbuf(pb);
-
-		gtk_icon_factory_add(factory, items[i].stock_id, icon_set);
-
-		gtk_icon_set_unref(icon_set);
-		g_object_unref(pb);
-	}
-	gtk_stock_add((GtkStockItem *) items, len);
-	gtk_icon_factory_add_default(factory);
-	g_object_unref(factory);
+	gtk_stock_add((GtkStockItem *) items, G_N_ELEMENTS(items));
 }
 
 
@@ -1978,13 +1900,16 @@ void ui_swap_sidebar_pos(void)
 	GtkWidget *pane = ui_lookup_widget(main_widgets.window, "hpaned1");
 	GtkWidget *left = gtk_paned_get_child1(GTK_PANED(pane));
 	GtkWidget *right = gtk_paned_get_child2(GTK_PANED(pane));
-	GtkWidget *box = ui_lookup_widget(main_widgets.window, "vbox1");
 
-	/* reparenting avoids scintilla problem with middle click paste */
-	gtk_widget_reparent(left, box);
-	gtk_widget_reparent(right, box);
-	gtk_widget_reparent(right, pane);
-	gtk_widget_reparent(left, pane);
+	g_object_ref(left);
+	g_object_ref(right);
+	gtk_container_remove (GTK_CONTAINER (pane), left);
+	gtk_container_remove (GTK_CONTAINER (pane), right);
+	/* only scintilla notebook should expand */
+	gtk_paned_pack1(GTK_PANED(pane), right, right == main_widgets.notebook, TRUE);
+	gtk_paned_pack2(GTK_PANED(pane), left, left == main_widgets.notebook, TRUE);
+	g_object_unref(left);
+	g_object_unref(right);
 
 	gtk_paned_set_position(GTK_PANED(pane), pane->allocation.width
 		- gtk_paned_get_position(GTK_PANED(pane)));
@@ -2089,8 +2014,163 @@ void ui_init_prefs(void)
 }
 
 
+/* Used to find out the name of the GtkBuilder retrieved object since
+ * some objects will be GTK_IS_BUILDABLE() and use the GtkBuildable
+ * 'name' property for that and those that don't implement GtkBuildable
+ * will have a "gtk-builder-name" stored in the GObject's data list. */
+static const gchar *ui_guess_object_name(GObject *obj)
+{
+	const gchar *name = NULL;
+
+	g_return_val_if_fail(G_IS_OBJECT(obj), NULL);
+
+	if (GTK_IS_BUILDABLE(obj))
+		name = gtk_buildable_get_name(GTK_BUILDABLE(obj));
+	if (! name)
+		name = g_object_get_data(obj, "gtk-builder-name");
+	if (! name)
+		return NULL;
+
+	return name;
+}
+
+
+/* Compatibility functions */
+GtkWidget *create_edit_menu1(void)
+{
+	return edit_menu1;
+}
+
+
+GtkWidget *create_prefs_dialog(void)
+{
+	return prefs_dialog;
+}
+
+
+GtkWidget *create_project_dialog(void)
+{
+	return project_dialog;
+}
+
+
+GtkWidget *create_toolbar_popup_menu1(void)
+{
+	return toolbar_popup_menu1;
+}
+
+
+GtkWidget *create_window1(void)
+{
+	return window1;
+}
+
+
+static GtkWidget *ui_get_top_parent(GtkWidget *widget)
+{
+	GtkWidget *parent;
+
+	g_return_val_if_fail(GTK_IS_WIDGET(widget), NULL);
+
+	for (;;)
+	{
+		if (GTK_IS_MENU(widget))
+			parent = gtk_menu_get_attach_widget(GTK_MENU(widget));
+		else
+			parent = gtk_widget_get_parent(widget);
+		if (parent == NULL)
+			parent = (GtkWidget*) g_object_get_data(G_OBJECT(widget), "GladeParentKey");
+		if (parent == NULL)
+			break;
+		widget = parent;
+	}
+
+	return widget;
+}
+
+
+void ui_init_builder(void)
+{
+	gchar *interface_file;
+	const gchar *name;
+	GError *error;
+	GSList *iter, *all_objects;
+	GtkWidget *widget, *toplevel;
+
+	/* prevent function from being called twice */
+	if (GTK_IS_BUILDER(builder))
+		return;
+
+	builder = gtk_builder_new();
+
+	gtk_builder_set_translation_domain(builder, GETTEXT_PACKAGE);
+
+	error = NULL;
+	interface_file = g_build_filename(app->datadir, "geany.glade", NULL);
+	if (! gtk_builder_add_from_file(builder, interface_file, &error))
+	{
+		/* Show the user this message so they know WTF happened */
+		dialogs_show_msgbox_with_secondary(GTK_MESSAGE_ERROR,
+			_("Geany cannot start!"), error->message);
+		/* Aborts */
+		g_error("Cannot create user-interface: %s", error->message);
+		g_error_free(error);
+		g_free(interface_file);
+		g_object_unref(builder);
+		return;
+	}
+	g_free(interface_file);
+
+	gtk_builder_connect_signals(builder, NULL);
+
+	edit_menu1 = GTK_WIDGET(gtk_builder_get_object(builder, "edit_menu1"));
+	prefs_dialog = GTK_WIDGET(gtk_builder_get_object(builder, "prefs_dialog"));
+	project_dialog = GTK_WIDGET(gtk_builder_get_object(builder, "project_dialog"));
+	toolbar_popup_menu1 = GTK_WIDGET(gtk_builder_get_object(builder, "toolbar_popup_menu1"));
+	window1 = GTK_WIDGET(gtk_builder_get_object(builder, "window1"));
+
+	g_object_set_data(G_OBJECT(edit_menu1), "edit_menu1", edit_menu1);
+	g_object_set_data(G_OBJECT(prefs_dialog), "prefs_dialog", prefs_dialog);
+	g_object_set_data(G_OBJECT(project_dialog), "project_dialog", project_dialog);
+	g_object_set_data(G_OBJECT(toolbar_popup_menu1), "toolbar_popup_menu1", toolbar_popup_menu1);
+	g_object_set_data(G_OBJECT(window1), "window1", window1);
+
+	all_objects = gtk_builder_get_objects(builder);
+	for (iter = all_objects; iter != NULL; iter = g_slist_next(iter))
+	{
+		if (! GTK_IS_WIDGET(iter->data))
+			continue;
+
+		widget = GTK_WIDGET(iter->data);
+
+		name = ui_guess_object_name(G_OBJECT(widget));
+		if (! name)
+		{
+			g_warning("Unable to get name from GtkBuilder object");
+			continue;
+		}
+
+		toplevel = ui_get_top_parent(widget);
+		if (toplevel)
+			ui_hookup_widget(toplevel, widget, name);
+	}
+	g_slist_free(all_objects);
+}
+
+
+static void init_custom_style(void)
+{
+	gchar *gtkrc_file = g_build_filename(app->datadir, "geany.gtkrc", NULL);
+
+	gtk_rc_parse(gtkrc_file);
+	g_free(gtkrc_file);
+}
+
+
 void ui_init(void)
 {
+	init_custom_style();
+
 	init_recent_files();
 
 	ui_widgets.statusbar = ui_lookup_widget(main_widgets.window, "statusbar");
@@ -2099,10 +2179,10 @@ void ui_init(void)
 	main_widgets.progressbar = progress_bar_create();
 
 	/* current word sensitive items */
-	widgets.popup_goto_items[0] = ui_lookup_widget(main_widgets.editor_menu, "goto_tag_definition1");
+	widgets.popup_goto_items[0] = ui_lookup_widget(main_widgets.editor_menu, "goto_tag_definition2");
 	widgets.popup_goto_items[1] = ui_lookup_widget(main_widgets.editor_menu, "context_action1");
-	widgets.popup_goto_items[2] = ui_lookup_widget(main_widgets.editor_menu, "find_usage1");
-	widgets.popup_goto_items[3] = ui_lookup_widget(main_widgets.editor_menu, "find_document_usage1");
+	widgets.popup_goto_items[2] = ui_lookup_widget(main_widgets.editor_menu, "find_usage2");
+	widgets.popup_goto_items[3] = ui_lookup_widget(main_widgets.editor_menu, "find_document_usage2");
 
 	widgets.popup_copy_items[0] = ui_lookup_widget(main_widgets.editor_menu, "cut1");
 	widgets.popup_copy_items[1] = ui_lookup_widget(main_widgets.editor_menu, "copy1");
@@ -2140,6 +2220,25 @@ void ui_init(void)
 }
 
 
+void ui_finalize_builder(void)
+{
+	if (GTK_IS_BUILDER(builder))
+		g_object_unref(builder);
+
+	/* cleanup refs lingering even after GtkBuilder is destroyed */
+	if (GTK_IS_WIDGET(edit_menu1))
+		gtk_widget_destroy(edit_menu1);
+	if (GTK_IS_WIDGET(prefs_dialog))
+		gtk_widget_destroy(prefs_dialog);
+	if (GTK_IS_WIDGET(project_dialog))
+		gtk_widget_destroy(project_dialog);
+	if (GTK_IS_WIDGET(toolbar_popup_menu1))
+		gtk_widget_destroy(toolbar_popup_menu1);
+	if (GTK_IS_WIDGET(window1))
+		gtk_widget_destroy(window1);
+}
+
+
 void ui_finalize(void)
 {
 	g_free(statusbar_template);
@@ -2148,10 +2247,15 @@ void ui_finalize(void)
 
 static void auto_separator_update(GeanyAutoSeparator *autosep)
 {
-	g_return_if_fail(autosep->ref_count >= 0);
+	g_return_if_fail(autosep->item_count >= 0);
 
 	if (autosep->widget)
-		ui_widget_show_hide(autosep->widget, autosep->ref_count > 0);
+	{
+		if (autosep->item_count > 0)
+			ui_widget_show_hide(autosep->widget, autosep->show_count > 0);
+		else
+			gtk_widget_destroy(autosep->widget);
+	}
 }
 
 
@@ -2160,9 +2264,9 @@ static void on_auto_separator_item_show_hide(GtkWidget *widget, gpointer user_da
 	GeanyAutoSeparator *autosep = user_data;
 
 	if (GTK_WIDGET_VISIBLE(widget))
-		autosep->ref_count++;
+		autosep->show_count++;
 	else
-		autosep->ref_count--;
+		autosep->show_count--;
 	auto_separator_update(autosep);
 }
 
@@ -2171,10 +2275,12 @@ static void on_auto_separator_item_destroy(GtkWidget *widget, gpointer user_data
 {
 	GeanyAutoSeparator *autosep = user_data;
 
+	autosep->item_count--;
+	autosep->item_count = MAX(autosep->item_count, 0);
 	/* GTK_WIDGET_VISIBLE won't work now the widget is being destroyed,
 	 * so assume widget was visible */
-	autosep->ref_count--;
-	autosep->ref_count = MAX(autosep->ref_count, 0);
+	autosep->show_count--;
+	autosep->show_count = MAX(autosep->item_count, 0);
 	auto_separator_update(autosep);
 }
 
@@ -2186,15 +2292,16 @@ static void on_auto_separator_item_destroy(GtkWidget *widget, gpointer user_data
 void ui_auto_separator_add_ref(GeanyAutoSeparator *autosep, GtkWidget *item)
 {
 	/* set widget ptr NULL when widget destroyed */
-	if (autosep->ref_count == 0)
+	if (autosep->item_count == 0)
 		g_signal_connect(autosep->widget, "destroy",
 			G_CALLBACK(gtk_widget_destroyed), &autosep->widget);
 
 	if (GTK_WIDGET_VISIBLE(item))
-	{
-		autosep->ref_count++;
-		auto_separator_update(autosep);
-	}
+		autosep->show_count++;
+
+	autosep->item_count++;
+	auto_separator_update(autosep);
+
 	g_signal_connect(item, "show", G_CALLBACK(on_auto_separator_item_show_hide), autosep);
 	g_signal_connect(item, "hide", G_CALLBACK(on_auto_separator_item_show_hide), autosep);
 	g_signal_connect(item, "destroy", G_CALLBACK(on_auto_separator_item_destroy), autosep);
@@ -2239,7 +2346,7 @@ GtkWidget *ui_lookup_widget(GtkWidget *widget, const gchar *widget_name)
 		if (GTK_IS_MENU(widget))
 			parent = gtk_menu_get_attach_widget(GTK_MENU(widget));
 		else
-			parent = widget->parent;
+			parent = gtk_widget_get_parent(widget);
 		if (parent == NULL)
 			parent = (GtkWidget*) g_object_get_data(G_OBJECT(widget), "GladeParentKey");
 		if (parent == NULL)
@@ -2254,8 +2361,17 @@ GtkWidget *ui_lookup_widget(GtkWidget *widget, const gchar *widget_name)
 }
 
 
+/* wraps gtk_builder_get_object()
+ * unlike ui_lookup_widget(), it does only support getting object created from the main
+ * UI file, but it can fetch any object, not only widgets */
+gpointer ui_builder_get_object (const gchar *name)
+{
+	return gtk_builder_get_object (builder, name);
+}
+
+
 /* Progress Bar */
-static guint progress_bar_timer_id = (guint) -1;
+static guint progress_bar_timer_id = 0;
 
 
 static GtkWidget *progress_bar_create(void)
@@ -2298,7 +2414,7 @@ static gboolean progress_bar_pulse(gpointer data)
  **/
 void ui_progress_bar_start(const gchar *text)
 {
-	g_return_if_fail(progress_bar_timer_id == (guint) -1);
+	g_return_if_fail(progress_bar_timer_id == 0);
 
 	if (! interface_prefs.statusbar_visible)
 		return;
@@ -2319,10 +2435,10 @@ void ui_progress_bar_stop(void)
 {
 	gtk_widget_hide(GTK_WIDGET(main_widgets.progressbar));
 
-	if (progress_bar_timer_id != (guint) -1)
+	if (progress_bar_timer_id != 0)
 	{
 		g_source_remove(progress_bar_timer_id);
-		progress_bar_timer_id = (guint) -1;
+		progress_bar_timer_id = 0;
 	}
 }
 
@@ -2407,7 +2523,7 @@ void ui_menu_add_document_items(GtkMenu *menu, GeanyDocument *active, GCallback
  * @a compare_func might be NULL to not sort the documents in the menu. In this case,
  * the order of the document tabs is used.
  *
- * See document_sort_by_display_name() for an example sort function.
+ * See document_compare_by_display_name() for an example sort function.
  *
  * @param menu Menu.
  * @param active Which document to highlight, or @c NULL.
@@ -2420,13 +2536,12 @@ void ui_menu_add_document_items_sorted(GtkMenu *menu, GeanyDocument *active,
 	GCallback callback, GCompareFunc compare_func)
 {
 	GtkWidget *menu_item, *menu_item_label, *image;
-	const GdkColor *color;
 	GeanyDocument *doc;
 	guint i, len;
 	gchar *base_name, *label;
 	GPtrArray *sorted_documents;
 
-	len = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
+	len = (guint) gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
 
 	sorted_documents = g_ptr_array_sized_new(len);
 	/* copy the documents_array into the new one */
@@ -2453,10 +2568,8 @@ void ui_menu_add_document_items_sorted(GtkMenu *menu, GeanyDocument *active,
 		gtk_container_add(GTK_CONTAINER(menu), menu_item);
 		g_signal_connect(menu_item, "activate", callback, doc);
 
-		color = document_get_status_color(doc);
 		menu_item_label = gtk_bin_get_child(GTK_BIN(menu_item));
-		gtk_widget_modify_fg(menu_item_label, GTK_STATE_NORMAL, color);
-		gtk_widget_modify_fg(menu_item_label, GTK_STATE_ACTIVE, color);
+		gtk_widget_set_name(menu_item_label, document_get_status_widget_class(doc));
 
 		if (doc == active)
 		{
@@ -2511,7 +2624,7 @@ void ui_editable_insert_text_callback(GtkEditable *editable, gchar *new_text,
 	gint i;
 
 	if (new_text_len == -1)
-		new_text_len = strlen(new_text);
+		new_text_len = (gint) strlen(new_text);
 
 	for (i = 0; i < new_text_len; i++, new_text++)
 	{
@@ -2529,7 +2642,6 @@ void ui_editable_insert_text_callback(GtkEditable *editable, gchar *new_text,
 GdkPixbuf *ui_get_mime_icon(const gchar *mime_type, GtkIconSize size)
 {
 	GdkPixbuf *icon = NULL;
-#if GTK_CHECK_VERSION(2, 14, 0)
 	gchar *ctype;
 	GIcon *gicon;
 	GtkIconInfo *info;
@@ -2555,7 +2667,7 @@ GdkPixbuf *ui_get_mime_icon(const gchar *mime_type, GtkIconSize size)
 			gtk_icon_info_free(info);
 		}
 	}
-#endif
+
 	/* fallback for builds with GIO < 2.18 or if icon lookup failed, like it might happen on Windows */
 	if (icon == NULL)
 	{
@@ -2584,3 +2696,19 @@ void ui_focus_current_document(void)
 	if (doc != NULL)
 		document_grab_focus(doc);
 }
+
+
+/** Finds the label text associated with stock_id
+ * @param stock_id stock_id to lookup e.g. @c GTK_STOCK_OPEN.
+ * @return The label text for stock
+ * @since Geany 1.22 */
+const gchar *ui_lookup_stock_label(const gchar *stock_id)
+{
+	GtkStockItem item;
+
+	if (gtk_stock_lookup(stock_id, &item))
+		return item.label;
+
+	g_warning("No stock id '%s'!", stock_id);
+	return NULL;
+}
